from typing import List
from haystack import component, Pipeline

import torch
from sentence_transformers import SentenceTransformer
from sentence_transformers.evaluation import (
    InformationRetrievalEvaluator,
    SequentialEvaluator,
)
from sentence_transformers.util import cos_sim

from datasets import Dataset
from sentence_transformers import SentenceTransformerModelCardData, SentenceTransformer
import time
from datetime import datetime
from sentence_transformers.losses import MatryoshkaLoss, TripletLoss

from sentence_transformers import SentenceTransformerTrainingArguments
from sentence_transformers.training_args import BatchSamplers

from sentence_transformers import SentenceTransformerTrainer
from datasets import load_dataset, concatenate_datasets


def prepare_matryoshka_evaluator(train_data_path:str, test_data_path:str, matryoshka_dimensions:List[int]) -> SequentialEvaluator:
    test_dataset = load_dataset("json", data_files="test_dataset.json", split="train")
    train_dataset = load_dataset("json", data_files="train_dataset.json", split="train")
    return prepare_matryoshka_evaluator(train_dataset, test_dataset, matryoshka_dimensions)

def prepare_matryoshka_evaluator(train_dataset:Dataset, test_dataset:Dataset, matryoshka_dimensions:List[int]) -> SequentialEvaluator:
    #test_dataset = load_dataset("json", data_files="test_dataset.json", split="train")
    #train_dataset = load_dataset("json", data_files="train_dataset.json", split="train")
    corpus_dataset = concatenate_datasets([train_dataset, test_dataset]) #only needed for evaluation
    # Convert the datasets to dictionaries
    corpus = dict(
        zip(corpus_dataset["id"], corpus_dataset["positive"])
    )  # Our corpus (cid => document)
    queries = dict(
        zip(test_dataset["id"], test_dataset["anchor"])
    )  # Our queries (qid => question)
    
    # Create a mapping of relevant document (1 in our case) for each query
    relevant_docs = {}  # Query ID to relevant documents (qid => set([relevant_cids])
    for q_id in queries:
        relevant_docs[q_id] = [q_id] #[corpus[q_id]] 
    
    matryoshka_evaluators = []
    # Iterate over the different dimensions
    for dim in matryoshka_dimensions:
        ir_evaluator = InformationRetrievalEvaluator(
            queries=queries,
            corpus=corpus,
            relevant_docs=relevant_docs,
            name=f"dim_{dim}",
            truncate_dim=dim,  # Truncate the embeddings to a certain dimension
            score_functions={"cosine": cos_sim},
        )
        matryoshka_evaluators.append(ir_evaluator)
    
    # Create a sequential evaluator
    evaluator = SequentialEvaluator(matryoshka_evaluators)
    return evaluator


def doMatryoshkaEval(evaluator: SequentialEvaluator, model: SentenceTransformer) -> None:
    results = evaluator(model)
    print(results)



@component
class EmbeddingModelMatryoshkaTripletEvaluator:
  """
  A component performing Matryoshka evaluation against the given embedding model.
  """
  @component.output_types()
  def run(self, evaluator: SequentialEvaluator, model: SentenceTransformer):
    print("[EmbeddingModelMatryoshkaTripletEvaluator]")
    print("started")
    doMatryoshkaEval(evaluator, model)
    print("completed")
    print("[EmbeddingModelMatryoshkaTripletEvaluator]")
    #return {"model_output_path": model_output_path}


# test
matryoshka_dimensions = [768, 512, 256, 128, 64]
train_dataset = load_dataset("json", data_files="train_dataset.json", split="train")
test_dataset = load_dataset("json", data_files="test_dataset.json", split="train")
evaluator = prepare_matryoshka_evaluator(train_dataset, test_dataset, matryoshka_dimensions)

model = SentenceTransformer(
    "BAAI/bge-base-en-v1.5",
    #model_kwargs={"attn_implementation": "sdpa"},  # sdpa will be used by default if available
    model_card_data=SentenceTransformerModelCardData(
        language="en",
        license="apache-2.0",
        model_name="BGE base ArgillaSDK Matryoshka",
    ),
)

evaluatortest= EmbeddingModelMatryoshkaTripletEvaluator()
evaluatortest.run(evaluator, model)



from typing import Any, Dict, List
from haystack import component, default_from_dict, default_to_dict, Document
from creao.core.Generator import *
import concurrent.futures
from creao.core.Endpoints import CreaoLLM


@component
class MapInterestToQuestionType:
    """
    A component mapping interest to question type
    """
    def __init__(self, type_of_questions:Dict[str,str], service:str = "default", pipeline_id:str = "default"):
        self.service = service
        self.pipeline_id = pipeline_id
        self.type_of_questions = type_of_questions
        if self.service == "default":
            self.llm = CreaoLLM(bot_name="question type extraction assistant", bot_content="You are given a list of interests and a list of types of questions. Your task is to identify the types of questions that can be asked based on the given interests. Reflective Thinking: Before you generate the final list of types of questions, pause to examine your assumptions, biases, and the mental models that the interests might use. Consider how the interests influence the types of questions that can be asked, and how they prioritize the types of questions. Be open to learning from previous similar tasks and improving the outcome. Explicitly share your reflective thought process before giving the final answer.")
        elif self.service == "openai":
            self.generator = Generator()
    @component.output_types(question_mapping=List[Dict[str, str]])
    def run(self, documents:List[Document]):
        """
        Map interest to question type
        :param interest: The interest to map
        :return: The question type
        """
        #print("mapping interest to question type with minimax")
        if len(documents) == 0:
            return {"question_mapping":[]}
        file_name = documents[0].meta["file_name"]
        chunk = documents[0].meta["chunk"]
        interests = [doc.content for doc in documents]
        def process_interest(interest):
            if self.service == "default":
                prompt = extract_compatible_question_type_prompt.format(
                    interest=interest,
                    types="\n".join(list(self.type_of_questions.values())),
                    file_name=file_name,
                    passage=chunk,
                )
                json_schema = {
                    "reasoning": {
                        "type": "string"
                    },
                    "list_of_extractable_types_of_questions": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }
                }
                response_json = self.llm.invoke(prompt,json_schema,"MapInterestToQuestionType", self.pipeline_id)
                try:
                    raw_answer = json.loads(response_json["reply"])
                except Exception as e:
                    print(f"MapInterestToQuestionType json decode error:{e}, with response_json:{response_json}")
                    raw_answer = {"list_of_extractable_types_of_questions":[]}
                if "list_of_extractable_types_of_questions" in raw_answer:
                    mapping = raw_answer['list_of_extractable_types_of_questions']
                else:
                    mapping = []
            elif self.service == "openai":
                mapping = self.generator.extract_compatible_question_type(interest, list(self.type_of_questions.values()), file_name, chunk)['list_of_extractable_types_of_questions']
            return {"interest": interest, "q_type": [m.lower() for m in mapping]}

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            question_mapping = list(executor.map(process_interest, interests))
        #print("mapping interest to question type with minimax done")
        return {"question_mapping": question_mapping}
        
    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(
            self,
            type_of_questions=self.type_of_questions,
            service=self.service,
            pipeline_id = self.pipeline_id
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "MapInterestToQuestionType":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)


from datasets import Dataset
import requests
import os
import tempfile
import uuid


def get_presigned_url(file_name:str, bucket_name:str="creaodev"):
    url = 'https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver'
    api_key= os.environ["CREAO_API_KEY"]
    headers = {"Content-Type":"application/json", "x-api-key":api_key}
    # Payload to be sent to the Lambda function via API Gateway
    payload = {
        "action": "s3",
        "file_name": file_name,
        "bucket_name": bucket_name
    }
    # Send the request
    response = requests.post(url,headers=headers, json=payload)
    #print("reponse:", response.content)
    return response.json()['url']

def submit_pipeline_job(pipeline_id:str, pipe_yaml:str, run_config_template:str, dataset:Dataset, username:str):
    # generate a random file name for the dataset based on uuid
    file_uuid = str(uuid.uuid4())
    # save the dataset to a temporary file
    # Use tempfile to create a temporary file
    with tempfile.NamedTemporaryFile(suffix=f"{file_uuid}.parquet", delete=False) as tmp_file:
        temp_file_path = tmp_file.name
        # Save the dataset as a Parquet file
        dataset.to_parquet(temp_file_path)
        print(f"Dataset saved to temporary file: {temp_file_path}")


    # get the presigned url for the dataset
    presigned_url = get_presigned_url(temp_file_path)

    # Open the zip file and upload it using the pre-signed URL
    with open(temp_file_path, 'rb') as f:
        upload_response = requests.put(presigned_url, data=f)

    # Check the response from the upload
    if upload_response.status_code == 200:
        print("File uploaded successfully")
    else:
        print("Upload failed with status code:", upload_response.status_code, upload_response.text)
    print("final clean up")
    # (Optional) Clean up the temporary file when done
    if os.path.exists(temp_file_path):
        os.remove(temp_file_path)
        print(f"Temporary file {temp_file_path} has been removed.")

